//  chud.cpp
//  Calculate Pi using the Chudnovsky Brothers' Binary Splitting Algorithm
//	Author: Alan Pittman (2002/03/03)	<apittman@rogers.com>

// Saturday, March 09, 2002

#include <iostream.h>
#include <iomanip.h>
#include <math.h>
#include <time.h>
#include <stdio.h>
#include "bigint.h"
#include "tuning.h"

extern "C" short *ivector(long n);
extern "C" void  free_ivector(short *v);
extern "C" void  fullmul(short w[],short u[],short v[],int n,int m,int flags);
extern "C" void  mpsqrt(short y[],short x[],short a[],int n);
extern "C" void  mpinv(short x[],short a[],int n);
extern "C" void  error(char error_text[]);

#if DEBUG
void show(char *c, short *u,int n);
#endif

#define NOP 0
#define LSH 2
#define STO 4
#define RCL 8

BigInt A, B, J, one, two, five, six;
int maxk, currk, oldpct = 0;
int prec=0, base=10;

////////////////////////////////////////////////////////////////////////////////
//  Pi = invroot (invroot (BigInt (640320, destprec), 2) * T, 1) * 53360 * Q
////////////////////////////////////////////////////////////////////////////////

void Pi (BigInt *T, BigInt *Q)
{
    int i,j,k,n;
    short *a,*x,*y;
    char str[80];
    FILE *fp1;

    k=(prec+3)/4;
	n=1; while (n<k) n<<=1;
	
    a=ivector(n);
    x=ivector(n);
    y=ivector(n);

    // a = 640320.0
    for (i=0;i<n;i++) a[i]=0;
    a[0]=0; a[1]=64; a[2]=320;

    // x = 1 / sqrt(a)
    mpsqrt(y,x,a,n);

    // a = T
    for (i=0;i<n;i++) a[i]=0;
    int tlen=(*T).copy(a, n);

    // a = x * a
    fullmul(a,x,a,n,n,0);

    // x = 1 / a 
    mpinv(x,a,n);

    // a = 53360.0
    for (i=0;i<n;i++) a[i]=0;
    a[0]=5; a[1]=3360; a[2]=0;

    // x = x * a;
    fullmul(x,x,a,n,n,0);

    // a = Q
    for (i=0;i<n;i++) a[i]=0;
    int qlen=(*Q).copy(a, n);
	
	cout << "Size of Q = " << qlen << " integers" << endl;

    // x = x * a
    fullmul(x,x,a,n,n,0);

    if ((fp1=fopen("pi.txt","w"))==NULL) error("fopen");

    for (i=0;i<n && x[i]==0;i++);
    fprintf(fp1,"%d.\n",x[i++]);
    for (k=0;i<n;i++) {
        sprintf(str,"%04d",x[i]);
        for (j=0;j<4;j++) {
            k++;
            if (k%1000==0)
                fprintf(fp1,"%c\n\n",str[j]);
            else if (k%50==0)
                fprintf(fp1,"%c\n",str[j]);
            else if (k%5==0)
                fprintf(fp1,"%c ",str[j]);
            else
                fprintf(fp1,"%c",str[j]);
        }
    }
    fprintf(fp1,"\n");
    fclose(fp1);

    free_ivector(y);
    free_ivector(x);
    free_ivector(a);
}

////////////////////////////////////////////////////////////////////////////////
//  a(n) = A + n.B    where  A = 13591409,  B = 545140134
////////////////////////////////////////////////////////////////////////////////

BigInt a (int n)
{
    BigInt v = A + B * n;

    v.sign(1 - 2 * (n & 1));

    return v;
}

////////////////////////////////////////////////////////////////////////////////
//  p(n) = (6.n - 1) * (2.n - 1) * (6.n - 5)
////////////////////////////////////////////////////////////////////////////////

BigInt p (int n)
{
    BigInt f = n, sixf = six * f;

    if (!n)
        return one;
    else
        return (sixf - one) * (two * f - one) * (sixf - five);
}

////////////////////////////////////////////////////////////////////////////////
//  q(n) = n^3 * c^3 / 24    where C = 640320
////////////////////////////////////////////////////////////////////////////////

BigInt q (int n)
{
    BigInt f = n;

    if (!n)
        return one;
    else
        return J * f * f * f;
}

////////////////////////////////////////////////////////////////////////////////
//  Recursive Binary Split
////////////////////////////////////////////////////////////////////////////////

void r (int N1, int N2, BigInt *T, BigInt *Q, BigInt *P)
{
    int pct;

    switch (N2 - N1)
    {
        case 0:
        {
            int assert = (N1 != N2);

            break;
        }
        case 1:
        {
            BigInt p0 = p (N1);

            *T = a (N1) * p0;
            *Q = q (N1);
            if (P) *P = p0;

            currk += 1;
            break;
        }
#if NUM_CASES >= 2
        case 2:
        {
            BigInt p0  = p (N1);
            BigInt p01 = p0 * p (N1 + 1);
            BigInt q1  = q (N1 + 1);

            *T = q1 * a (N1) * p0 +
                 a (N1 + 1) * p01;
            *Q = q (N1) * q1;
            if (P) *P = p01;

            currk += 4;
            break;
        }
#endif
#if NUM_CASES >= 3
        case 3:
        {
            BigInt p0   = p (N1);
            BigInt p01  = p0 * p (N1 + 1);
            BigInt p012 = p01 * p (N1 + 2);
            BigInt q2   = q (N1 + 2);
            BigInt q12  = q (N1 + 1) * q2;

            *T = q12 * a (N1) * p0
                + q2 * a (N1 + 1) * p01
                + a (N1 + 2) * p012;
            *Q = q (N1) * q12;
            if (P) *P = p012;

            currk += 8;
            break;
        }
#endif
#if NUM_CASES >= 4
        case 4:
        {
            BigInt p0    = p (N1);
            BigInt p01   = p0 * p (N1 + 1);
            BigInt p012  = p01 * p (N1 + 2);
            BigInt p0123 = p012 * p (N1 + 3);
            BigInt q3    = q (N1 + 3);
            BigInt q23   = q (N1 + 2) * q3;
            BigInt q123  = q (N1 + 1) * q23;

            *T = q123 * a (N1) * p0
                + q23 * a (N1 + 1) * p01
                + q3 * a (N1 + 2) * p012
                + a (N1 + 3) * p0123;
            *Q = q (N1) * q123;
            if (P) *P = p0123;

            currk += 12;

            break;
        }
#endif
        default:
        {
            int Nm = N1 + (int)((double)(N2 - N1) * BINSPLIT);
            BigInt LT, LQ, LP, RT, RQ, RP;

            r (N1, Nm, &LT, &LQ, &LP);
            r (Nm, N2, &RT, &RQ, (P ? &RP : 0));

            *T = RQ * LT + LP * RT;
            *Q = LQ * RQ;
            if (P) *P = LP * RP;

            currk += N2 - N1;
        }
    }

    pct = (int) (100.0 * currk / maxk);

    if (pct != oldpct)
    {
        cerr << pct << "% complete\r";
        cerr.flush ();

        oldpct = pct;
    }
}

////////////////////////////////////////////////////////////////////////////////
//  Chudnovsky Algorithm
////////////////////////////////////////////////////////////////////////////////

void pib (int destprec)
{
    int n;
    time_t tt;
    clock_t tc;
    BigInt T, Q;

    cerr << "Using the Chudnovsky Brothers' Binary Splitting Algorithm" << endl;

    A = 13591409;
    B = 545140134;
    J = 101568000;
    J = J * 107701824;       // J = 10939058860032000

    one = 1;
    two = 2;
    five = 5;
    six = 6;

    n = (int) (destprec * log ((double) 10.0) / 32.65445004177);
    maxk = (int) (n * (log ((double) n) / log (2.0) + 1));
    currk = 0;

    tt = time (0);
    tc = clock ();

    r (0, n + 1, &T, &Q, 0);

    cerr.setf (ios::fixed, ios::floatfield);
    cerr << "100% complete, took " << setprecision (3)
		 << (clock () - tc) / (double) CLOCKS_PER_SEC
		 << " seconds CPU time, elapsed time " << setprecision (0)
		 << difftime (time (0), tt) << " seconds" << endl;

    cerr << "Final value" << endl;

    tt = time (0);
    tc = clock ();

//  a = invroot (invroot (BigInt (640320, destprec), 2) * T, 1) * 53360 * Q

    Pi (&T, &Q);

    cerr.setf (ios::fixed, ios::floatfield);
    cerr << " took " << setprecision (3)
		 << (clock () - tc) / (double) CLOCKS_PER_SEC
		 << " seconds CPU time, elapsed time " << setprecision (0)
		 << difftime (time (0), tt) << " seconds" << endl;
}

////////////////////////////////////////////////////////////////////////////////
//  Program Mainline
////////////////////////////////////////////////////////////////////////////////

int main(int argc, char *argv[])
{
    time_t tt;
    clock_t tc;
    char str[80];

	cout << "Version 1.4\n";
    if (argc > 1)
        sscanf(argv[1],"%ld",&prec);
    while (prec == 0) {
        printf ("enter the number of digits => ");
        scanf ("%s",str );
        sscanf (str,"%ld",&prec);
    }

    cerr << "Calculating Pi to " << prec << " digits" << endl;

    tt = time (0);
    tc = clock ();

    pib(prec);

    cerr.setf (ios::fixed, ios::floatfield);
    cerr << "Total CPU time " << setprecision (3)
		 << (clock () - tc) / (double) CLOCKS_PER_SEC
		 << " seconds, total elapsed time " << setprecision (0)
		 << difftime (time (0), tt) << " seconds" << endl;

    return 0;
}

#if DEBUG
void show(char *c, short *u,int n)
{
#if 1
    int i;
    printf("%s\n",c);
    for (i=0;i<n;i++) {
        if ((i+1)%16==0 && (i+1)!=n) {
            printf("%04d\n",u[i]);
        } else {
            printf("%04d ",u[i]);
        }
    }
    printf("\n\n");
#else
    int i;
    printf("%s\n",c);
    for (i=0;i<n;i++) {
        printf("%04d",u[i]);
    }
    printf("\n\n");
#endif
}
#endif
